/*
	$Id: objects.cpp,v 1.8 1998/12/31 16:43:26 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Game objects (pacman, ghosts, powerups)

*/

#include "../ClanLib/clanlib.h"
#include "pacman.h"
#include "objects.h"

/***************
 Moving object
***************/

GameObject_Moving::GameObject_Moving(int _x, int _y, PacManApplication *_app)
{
	app = _app;
	x = _x;
	y = _y;
	dest_x = _x;
	dest_y = _y;
	speed = 1.0;
}

bool GameObject_Moving::verify_move(int dx, int dy)
{
	if (app->map[dest_x + dx][dest_y + dy] < 4) return true;
	return false;
}

bool GameObject_Moving::do_move(int dx, int dy)
{
	if (verify_move(dx,dy) == false) return false;

	dest_x += dx;
	dest_y += dy;

	return true;
}

bool GameObject_Moving::move(float time_elapsed)
{
        bool big_move = true;

	do
	{
	        big_move = true;
	        float delta_x = dest_x-x;

		if (fabs(delta_x) < time_elapsed*speed)
		{
		        x = dest_x;
		}
		else if (delta_x>0)
		{
		        x += time_elapsed*speed;
		}
		else
		{
		        x -= time_elapsed*speed;
		}

		float delta_y = dest_y-y;
		if (fabs(delta_y) < time_elapsed*speed)
		{
		        y = dest_y;
	        }
		else if (delta_y>0)
		{
		        y += time_elapsed*speed;
		}
		else
		{
		        y -= time_elapsed*speed;
		}

	        if (fabs(delta_x) < time_elapsed*speed &&
		    fabs(delta_y) < time_elapsed*speed)
		{
		        if (delta_x < delta_y) time_elapsed -= fabs(delta_y)/speed;
			        else time_elapsed -= fabs(delta_x)/speed;
			big_move = !event_reached_dest();
		}
	} 
	while (!big_move);

	return true;
}


/***************
 Pacman object
****************/

GameObject_Pacman::GameObject_Pacman(int x, int y, PacManApplication *app)
: GameObject_Moving(x, y, app)
{
	move_dir = 1;
	wanted_dir = -1;
	anim_pos = 0;
	anim_dir = 1;

	i_am_dead = false;
	got_powerup = false;

	set_speed(4.0);
}

void GameObject_Pacman::show(int center_x, int center_y)
{
	int width = app->sur_maptiles->get_provider()->get_width()-6;
	int height = app->sur_maptiles->get_provider()->get_height()-6;
	int anim_length = app->sur_pacman->get_provider()->get_no_sprs() / 4;

	app->sur_pacman->put_screen(
		(int) (x*width)-center_x,
		(int) (y*height)-center_y,
		anim_pos+anim_length*move_dir);
}

bool GameObject_Pacman::hit_check(float hit_x, float hit_y)
{
	if (fabs(x-hit_x) < 0.25 && fabs(y-hit_y) < 0.25) return true;
	return false;
}

bool GameObject_Pacman::turn(float time_elapsed)
{
	if (got_powerup && CL_System::get_time() - powerup_starttime>20000)
	{
		got_powerup = false;
	}

	if (i_am_dead) 
	{
		app->score -= 5000;
		if (app->score < 0) app->score = 0;
		return false;
	}	

	int anim_length = app->sur_pacman->get_provider()->get_no_sprs() / 4;

	anim_pos += anim_dir;
	if (anim_pos >= anim_length-1 || anim_pos == 0) anim_dir*=-1;

	switch (move_dir)
	{
	case 0:
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_RIGHT))
		{
			move_dir = 1;
			wanted_dir = -1;
			do_move(1,0);
		}
		break;

	case 1:
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_LEFT))
		{
			move_dir = 0;
			wanted_dir = -1;
			do_move(-1,0);
		}
		break;

	case 2:
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_DOWN))
		{
			move_dir = 3;
			wanted_dir = -1;
			do_move(0,1);
		}
		break;

	case 3:
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_UP))
		{
			move_dir = 2;
			wanted_dir = -1;
			do_move(0,-1);
		}
		break;
	}


	if (move_dir < 2)
	{
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_UP))
			wanted_dir = 2;

		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_DOWN))
			wanted_dir = 3;
	}
	else
	{
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_LEFT))
			wanted_dir = 0;

		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_RIGHT))
			wanted_dir = 1;
	}


	return move(time_elapsed);
}

bool GameObject_Pacman::event_reached_dest()
{
	if (app->map[dest_x][dest_y] == 1) // std
	{
		app->score += 125;
		app->nof_eggs--;
		app->sfx_namnam->play();
	}
	if (app->map[dest_x][dest_y] == 2) // powerup
	{
		app->score += 500;
		app->nof_eggs--;
		got_powerup = true;
		powerup_starttime = CL_System::get_time();
//		app->sfx_namnam->play();
	}
	app->map[dest_x][dest_y] = 3;

	if (move_dir < 2)
	{
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_UP) && verify_move(0, -1))
			move_dir = 2;

		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_DOWN) && verify_move(0, 1))
			move_dir = 3;
	}
	else
	{
		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_LEFT) && verify_move(-1, 0))
			move_dir = 0;

		if (CL_Keyboard::get_keycode(CL_Keyboard::KEY_RIGHT) && verify_move(1, 0))
			move_dir = 1;
	}

	
	switch (wanted_dir)
	{
	case -1: break;

	case 0:
	        if (verify_move(-1, 0)) 
		{
		        move_dir = 0;
			wanted_dir = -1;
                }
		break;

	case 1:
		if (verify_move(1, 0)) 
		{
		        move_dir = 1;
			wanted_dir = -1;
                }
		break;

	case 2:
		if (verify_move(0, -1)) 
		{
		        move_dir = 2;
			wanted_dir = -1;
                }
		break;

	case 3:
		if (verify_move(0, 1)) 
		{
		        move_dir = 3;
			wanted_dir = -1;
                }
		break;
	}


	switch (move_dir)
	{
	case 0:
		do_move(-1, 0);
		break;

	case 1:
		do_move(1, 0);
		break;

	case 2:
		do_move(0, -1);
		break;

	case 3:
		do_move(0, 1);
		break;
	}

	if (x == dest_x && y == dest_y) return false; // Damn - we're facing a wall!

	return true;
}

/**************
 Ghost object
**************/

GameObject_Ghost::GameObject_Ghost(int x, int y, PacManApplication *app)
: GameObject_Moving(x, y, app)
{
	move_dir = rand()%4;
	
	red_ghost = false;

	int anim_length = app->sur_ghost->get_provider()->get_no_sprs() / 2;

	anim_pos = rand()%anim_length;
	anim_dir = rand()%2;
	if (anim_dir == 0) anim_dir = -1;

	if (anim_pos == 0) anim_dir = 1;
	if (anim_pos == anim_length-1) anim_dir = -1;

	set_speed(3.0);
}

void GameObject_Ghost::show(int center_x, int center_y)
{
	int width = app->sur_maptiles->get_provider()->get_width()-6;
	int height = app->sur_maptiles->get_provider()->get_height()-6;
	
	if (app->player == NULL)
	{
		app->sur_ghost->put_screen(
			(int) (x*width)-center_x,
			(int) (y*height)-center_y,
			anim_pos);

		return;
	}

	if ((app->player->get_got_powerup() && int(x+y)%2==0) ||
		(!app->player->get_got_powerup() && red_ghost))
		app->sur_ghost->put_screen(
			(int) (x*width)-center_x,
			(int) (y*height)-center_y,
			anim_pos+6);
	else
		app->sur_ghost->put_screen(
			(int) (x*width)-center_x,
			(int) (y*height)-center_y,
			anim_pos);
}

bool GameObject_Ghost::turn(float time_elapsed)
{
	int anim_length = app->sur_ghost->get_provider()->get_no_sprs() / 2;

	anim_pos += anim_dir;
	if (anim_pos == 0 || anim_pos == anim_length-1) anim_dir *= -1;

	// pacman hitchecking:
	if (app->player != NULL)
	{
		if (app->player->hit_check(x, y)) // HA! Got player... nam nam!
		{
			if (app->player->get_got_powerup()) // ARRRG! He's got a powerup!
			{
				app->score += 1000;
				return false;
			}
			else
				app->player->you_are_dead();
		}
	}

	return move(time_elapsed);
}

bool GameObject_Ghost::event_reached_dest()
{
	int try_dirs_x[3];
	int try_dirs_y[3];
	int try_dirs_id[3];

	switch (move_dir)
	{
	case 0: // left
		try_dirs_x[0] = 0; try_dirs_y[0]=-1; try_dirs_id[0] = 2;
		try_dirs_x[1] =-1; try_dirs_y[1]= 0; try_dirs_id[1] = 0;
		try_dirs_x[2] = 0; try_dirs_y[2]= 1; try_dirs_id[2] = 3;
		break;

	case 1: // right
		try_dirs_x[0] = 0; try_dirs_y[0]=-1; try_dirs_id[0] = 2;
		try_dirs_x[1] = 1; try_dirs_y[1]= 0; try_dirs_id[1] = 1;
		try_dirs_x[2] = 0; try_dirs_y[2]= 1; try_dirs_id[2] = 3;
		break;

	case 2: // up
		try_dirs_x[0] =-1; try_dirs_y[0]= 0; try_dirs_id[0] = 0;
		try_dirs_x[1] = 0; try_dirs_y[1]=-1; try_dirs_id[1] = 2;
		try_dirs_x[2] = 1; try_dirs_y[2]= 0; try_dirs_id[2] = 1;
		break;

	case 3: // down
		try_dirs_x[0] =-1; try_dirs_y[0]= 0; try_dirs_id[0] = 0;
		try_dirs_x[1] = 0; try_dirs_y[1]= 1; try_dirs_id[1] = 3;
		try_dirs_x[2] = 1; try_dirs_y[2]= 0; try_dirs_id[2] = 1;
		break;
	}

	bool follow_trail = false;
	if (app->map[dest_x][dest_y] == 3)
	{
		// follow trail if any:
	        for (int i=0; i<3; i++)
		        if (app->map[dest_x+try_dirs_x[i]][dest_y+try_dirs_y[i]] == 3) 
			        if (app->player != NULL)
			                follow_trail = !app->player->get_got_powerup();
		// don't follow trail if player got power up or is dead
	}
	
//	if (follow_trail) cout << "ghost in follow trail mode" << endl;

	int dirs_avail = 0;
	for (int i=0; i<3; i++)
	{
		if (follow_trail)
		{
			if (app->map[dest_x+try_dirs_x[i]][dest_y+try_dirs_y[i]] == 3) dirs_avail++;
		}
		else if (verify_move(try_dirs_x[i], try_dirs_y[i])) dirs_avail++;
	}

	if (dirs_avail == 0) // Dead end - turn around
	{
		do_move(try_dirs_x[1]*(-1), try_dirs_y[1]*(-1));
		switch (move_dir)
		{
			case 0: move_dir=1; break;
			case 1: move_dir=0; break;
			case 2: move_dir=3; break;
			case 3: move_dir=2; break;
		}
	}
	else
	{
		int new_dir = rand()%dirs_avail;
		for (int i=0; i<3; i++)
		{
			if ((follow_trail && app->map[dest_x+try_dirs_x[i]][dest_y+try_dirs_y[i]] == 3) ||
				(!follow_trail && verify_move(try_dirs_x[i], try_dirs_y[i])))
			{
				if (new_dir == 0)
				{
					if (app->map[dest_x][dest_y] == 3)
					{
						app->map[dest_x][dest_y] = 0;
						red_ghost = true;
					}
					else
					{
						red_ghost = false;
					}

					do_move(try_dirs_x[i], try_dirs_y[i]);
					move_dir = try_dirs_id[i];
				}
				new_dir--;
			}
		}
	}

	return true;
}
