/*
	$Id: objects.h,v 1.4 1998/12/25 09:32:20 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Game objects (pacman, ghosts, powerups)


	$Log: objects.h,v $
	Revision 1.4  1998/12/25 09:32:20  mbn
	Totally new COOL version of the Pacman game. New features include:

	* Pacman leaves a trail behind it, making the ghosts more tricky to avoid.
	* Powerup in the corners.
	* Better transparency support.
	* Score decrease when not eating anything.
	* Random generated levels (but I'm not running srand() :-)).
	* Ghost changes color when chased or on your trail.
	* Fixed typo in the "Bonus Bananas" text.

	Now I only need hiscore support :-)

	Revision 1.3  1998/12/23 20:06:20  mbn
	Applied Danied Vogels patch to the pacman game.

	Revision 1.2  1998/12/21 17:56:58  mbn
	Tried to use the resource manager... but FAILED!

	Revision 1.1.1.1	1998/12/20 23:43:04	 mbn
	Pacman game moved to new repository.

	Revision 1.1.1.1	1998/08/23 23:36:20	 bbn
	I made this Pacman game to demonstrate some features in ClanLib.

	Revision 1.1	1998/05/06 22:20:51	 mbn
	Added small Pacman demonstration game to ClanLib.

*/

#ifndef header_objects
#define header_objects

class PacManApplication;

class GameObject
{
public:
	virtual ~GameObject() { return; }

	virtual float get_x()=0;
	virtual float get_y()=0;

	virtual void show(int center_x, int center_y)=0; // draw object on screen
	virtual bool turn(float time_elapsed)=0; // Moves the object around

	virtual bool hit_check(float x, float y)=0;
};

class GameObject_Moving : public GameObject
{
private:
	float speed;

protected:
	PacManApplication *app;
	float x, y;
	int dest_x, dest_y;

	bool move(float time_elapsed);
	bool verify_move(int dx, int dy);
	bool do_move(int dx, int dy);

	virtual bool event_reached_dest()=0;

public:
	GameObject_Moving(int x, int y, PacManApplication *app);
	virtual ~GameObject_Moving() { return; }

	virtual float get_x() { return x; }
	virtual float get_y() { return y; }

	void set_speed(float new_speed) { speed = new_speed; }
};

class GameObject_Pacman : public GameObject_Moving
{
protected:
	int anim_pos, anim_dir;
	int move_dir;
	int wanted_dir;
	bool i_am_dead;
	bool got_powerup;
	int powerup_starttime;

	virtual bool event_reached_dest();

public:
	GameObject_Pacman(int x, int y, PacManApplication *app);

	virtual void show(int center_x, int center_y);
	virtual bool turn(float time_elapsed);
	virtual bool hit_check(float x, float y);

	void you_are_dead() { i_am_dead = true; }
	bool get_got_powerup() { return got_powerup; }
};

class GameObject_Ghost : public GameObject_Moving
{
protected:
	int anim_pos, anim_dir;
	int move_dir;
	
	bool red_ghost;

	virtual bool event_reached_dest();

public:
	GameObject_Ghost(int x, int y, PacManApplication *app);

	virtual void show(int center_x, int center_y);
	virtual bool turn(float time_elapsed);
	virtual bool hit_check(float /*x*/, float /*y*/) { return false; }
};

#endif
