/*
	$Id: pacman.cpp,v 1.8 1998/12/30 21:00:48 mbn Exp $

	------------------------------------------------------------------------
	ClanLib, the platform independent game SDK.

	This library is distributed under the GNU LIBRARY GENERAL PUBLIC LICENSE
	version 2. See COPYING for details.

	For a total list of contributers see CREDITS.

	------------------------------------------------------------------------

	File purpose:
		Small Pacman game. Main file.

*/

#include <iostream.h>
#include <clanlib.h>
#include <clanlayer2.h>

#include "pacman.h"
#include "objects.h"

PacManApplication app;

char *PacManApplication::get_title()
{
	return "ClanLib Pacman game";
}

int PacManApplication::main(int /*argc*/, char** /*argv*/)
{
	cout << "ClanLib Pacman game, written by Magnus Norddahl" << endl;

	resources = new CL_ResourceManager("pacman.scr", "pacman.dat");

	CL_System::init_display();
	CL_System::init_sound();

	init_display();
	init_map(0);
	init_objects();

	run_game();

	deinit();
	delete resources;

	return 0;
}

void PacManApplication::init_display()
{
	// Set videomode resolution and bpp (bits per pixel)
	CL_Display::set_videomode(640, 480, 8);

	// Set palette
	CL_Palette *pal = CL_Res_Palette::load("Game/pal_game", *resources);
	CL_Display::set_palette(pal);
	delete pal;

	sur_maptiles = CL_Res_Surface::load("Game/sur_maptiles", *resources);
	sur_pacman = CL_Res_Surface::load("Game/sur_pacman", *resources);
	sur_ghost = CL_Res_Surface::load("Game/sur_ghost", *resources);
	sur_bar = CL_Res_Surface::load("Game/sur_bar", *resources);

	fnt_clansoft = CL_Res_Font::load("Fonts/fnt_clansoft", *resources);

	// Yeah I know, it has nothing to do with the display...
	sfx_namnam = CL_Res_Sample::load("Game/sfx_namnam", *resources);
}
/*
old map: read from a file

void PacManApplication::init_map(int level)
{
	int x,y;

	center_x = 0;
	center_y = 0;
	score = 0;

	// Load map from datafile

	CL_String level_name;
	level_name << "Levels/level" << (level+1); // eg. "level1"

	cout << "Using level: '" << level_name << "'" << endl;

	CL_InputSource *input = CL_Res_Raw::load(level_name, *resources);
	cl_assert(input != NULL);

	cout << "Reading level" << endl;

	for (y=0; y<MAP_HEIGHT; y++)
	{
		for (x=0; x<MAP_WIDTH; x++)
		{
			int cur_tile = input->read_char8();
			if (cur_tile == 32) map[x][y] = 1; else map[x][y] = 4;
		}
		input->read_char8(); // '\n' in text
//		input->read_char8(); // '\r' in text
	}

	delete input;

	// Adjust maptiles so neightbor walls will connect together

	for (y=0; y<MAP_HEIGHT; y++)
	{
		for (x=0; x<MAP_WIDTH; x++)
		{
			if (x>=7 && x<=11 && y>=6 && y<=8 && map[x][y]==1) map[x][y] = 0;

			if (map[x][y] <= 1) continue;

			int new_image = 0;
			if (x>0	 && map[x-1][y]>1) new_image |= 1; // left flag
			if (x<MAP_WIDTH-1 && map[x+1][y]>1) new_image |= 2; // right flag
			if (y>0	 && map[x][y-1]>1) new_image |= 4; // up flag
			if (y<MAP_HEIGHT-1 && map[x][y+1]>1) new_image |= 8; // down flag

			map[x][y] = new_image+4;
		}
	}
}
*/

// new map: random generate
#define NUM_CONNECTS 50
void PacManApplication::init_map(int level)
{
	int x,y;

	center_x = 0;
	center_y = 0;
	score = 0;

	// 1. clear map with walls and gaps
	for (y=0; y<MAP_HEIGHT; y++)
		for (x=0;x<MAP_WIDTH; x++)
			map[x][y] = 4;

	// 2. build tunnels in level
	map[MAP_WIDTH/2][MAP_HEIGHT/2] = 0;
	map[MAP_WIDTH/2-1][MAP_HEIGHT/2] = 1;
	map[MAP_WIDTH/2+1][MAP_HEIGHT/2] = 1;
	map[MAP_WIDTH/2][MAP_HEIGHT/2-1] = 1;
	map[MAP_WIDTH/2][MAP_HEIGHT/2+1] = 1;
	
	int tunnel_x[4][MAP_WIDTH*MAP_HEIGHT];
	int tunnel_y[4][MAP_WIDTH*MAP_HEIGHT];

	tunnel_x[0][0] = MAP_WIDTH/2;
	tunnel_y[0][0] = MAP_HEIGHT/2-2;
	tunnel_x[1][0] = MAP_WIDTH/2;
	tunnel_y[1][0] = MAP_HEIGHT/2+2;
	tunnel_x[2][0] = MAP_WIDTH/2-2;
	tunnel_y[2][0] = MAP_HEIGHT/2;
	tunnel_x[3][0] = MAP_WIDTH/2+2;
	tunnel_y[3][0] = MAP_HEIGHT/2;

	for (int pos=1; pos<MAP_WIDTH*MAP_HEIGHT; pos++)
	{
		int filled_level = 0;
		for (int tun=0; tun<4; tun++)
		{
			if (pos == 1)
				map[tunnel_x[tun][pos-1]][tunnel_y[tun][pos-1]] = 1;

			if (find_neightbor(
				tunnel_x[tun][pos-1],
				tunnel_y[tun][pos-1],
				&tunnel_x[tun][pos],
				&tunnel_y[tun][pos])==false)
			{
				bool total_dead_end = true;
				for (int pos2=pos-2; pos2>=0; pos2--)
				{
					if (find_neightbor(
						tunnel_x[tun][pos2],
						tunnel_y[tun][pos2],
						&tunnel_x[tun][pos],
						&tunnel_y[tun][pos]))
					{
						total_dead_end = false;
						break;
					}
				}

				if (total_dead_end)
				{
					tunnel_x[tun][pos] = tunnel_x[tun][pos-1];
					tunnel_y[tun][pos] = tunnel_y[tun][pos-1];
					filled_level++;
				}
			}
			
			map[tunnel_x[tun][pos]][tunnel_y[tun][pos]] = 1;
		}
		
		if (filled_level == 4) break;
	}

	// 3. connect gaps random
	for (int i=0; i<NUM_CONNECTS; i++)
	{
		int x = rand()%(MAP_WIDTH/2-3);
		int y = rand()%(MAP_HEIGHT/2-3);
		
		x *= 2; y *= 2;
		x += 3; y += 3;
		
		switch (rand()%4)
		{
		case 0:
			map[x][y-1] = 1;
			break;
		case 1:
			map[x][y+1] = 1;
			break;
		case 2:
			map[x-1][y] = 1;
			break;
		case 3:
			map[x+1][y] = 1;
			break;
		}
	}

	for (y=0; y<MAP_HEIGHT; y++)
	{
		for (x=0; x<MAP_WIDTH; x++)
		{
//			if (x>=7 && x<=11 && y>=6 && y<=8 && map[x][y]==1) map[x][y] = 0;

			if (map[x][y] <= 1) continue;

			int new_image = 0;
			if (x>0	 && map[x-1][y]>3) new_image |= 1; // left flag
			if (x<MAP_WIDTH-1 && map[x+1][y]>3) new_image |= 2; // right flag
			if (y>0	 && map[x][y-1]>3) new_image |= 4; // up flag
			if (y<MAP_HEIGHT-1 && map[x][y+1]>3) new_image |= 8; // down flag

			map[x][y] = new_image+4;
		}
	}
	
	map[1][1] = 2;
	map[MAP_WIDTH-2][1] = 2;
	map[1][MAP_HEIGHT-2] = 2;
	map[MAP_WIDTH-2][MAP_HEIGHT-2] = 2;

       	nof_eggs = 0;	
	for (y=0; y<MAP_HEIGHT; y++)
                for (x=0; x<MAP_WIDTH; x++)       
			if ((map[x][y] == 1) || (map[x][y] == 2))
				nof_eggs++;
}

bool PacManApplication::find_neightbor(int x, int y, int *xx, int *yy)
{
	*xx = x;
	*yy = y;

	int num_dirs = 0;
	if (x>1 && map[x-2][y]>=4) num_dirs++;
	if (y>1 && map[x][y-2]>=4) num_dirs++;
	if (x<MAP_WIDTH-3 && map[x+2][y]>=4) num_dirs++;
	if (y<MAP_HEIGHT-3 && map[x][y+2]>=4) num_dirs++;
	
	int sel = rand()%(num_dirs+1);

	if (x>1 && map[x-2][y]>=4 && (sel--)==0) { map[x-1][y]=1; *xx = x-2; return true; }
	if (y>1 && map[x][y-2]>=4 && (sel--)==0) { map[x][y-1]=1; *yy = y-2; return true; }
	if (x<MAP_WIDTH-3 && map[x+2][y]>=4 && (sel--)==0) { map[x+1][y]=1; *xx = x+2; return true; }
	if (y<MAP_HEIGHT-3 && map[x][y+2]>=4 && (sel--)==0) { map[x][y+1]=1; *yy = y+2; return true; }

	return false;
}

void PacManApplication::init_objects()
{
	player = new GameObject_Pacman(MAP_WIDTH/2, MAP_HEIGHT/2, this);
	objects.add(player);

	int i;
	for (i=-2; i<=2; i++) objects.add(new GameObject_Ghost(1, 1, this));
	for (i=-2; i<=2; i++) objects.add(new GameObject_Ghost(MAP_WIDTH-2, 1, this));
	for (i=-2; i<=2; i++) objects.add(new GameObject_Ghost(1, MAP_HEIGHT-2, this));
	for (i=-2; i<=2; i++) objects.add(new GameObject_Ghost(MAP_WIDTH-2, MAP_HEIGHT-2, this));
}

void PacManApplication::run_game()
{
	int start_time = CL_System::get_time();
	int cur_frame = 0;

	int begin_time = CL_System::get_time();
	int score_time = CL_System::get_time();

	bool welcome_shown = false;
	while (CL_Keyboard::get_keycode(CL_Keyboard::KEY_ESCAPE) == false)
	{
		if (player == NULL && CL_Keyboard::get_keycode(CL_Keyboard::KEY_SPACE))
		{
			player = new GameObject_Pacman(MAP_WIDTH/2, MAP_HEIGHT/2, this);
			objects.add(player);

			if (blowups.get_num_items() < 10)
			{
				blowups.add(new FontBlowUp("YOU ARE DEAD", 320, 320, fnt_clansoft));
				blowups.add(new FontBlowUp("Press space to restart", 320, 360, fnt_clansoft));
			}
		}

		float time_elapsed = (CL_System::get_time() - begin_time)/(float) 1000;
		begin_time = CL_System::get_time();
		cur_frame++;
		CL_Display::clear_display();
		
		if (CL_System::get_time()-score_time > 1000 && player != NULL)
		{
			score -= 50;
			if (score < 0) score = 0;
			score_time = CL_System::get_time(); 
		}

//		sur_bar->put_screen(0, 0);

		int width = sur_maptiles->get_provider()->get_width()-6;
		int height = sur_maptiles->get_provider()->get_height()-6;

		if (player != NULL)
		{
			center_x = int((player->get_x()+0.5)*width-320);
			center_y = int((player->get_y()+0.5)*height-240);
			if (center_x < 0) center_x = 0;
			if (center_y < 0) center_y = 0;
			if (center_x > MAP_WIDTH*width-640) center_x = MAP_WIDTH*width-640;
			if (center_y > MAP_HEIGHT*height-480) center_y = MAP_HEIGHT*height-480;
		}

		int powerup_spr = ((CL_System::get_time()-start_time)/250)%2;

		for (int y=0; y<MAP_HEIGHT; y++)
		{
			for (int x=0; x<MAP_WIDTH; x++)
			{
				if (map[x][y] == 0) continue;
				int scr_x = x*width-6 - center_x;
				int scr_y = y*height-6 - center_y;
//				if (scr_x < 0 || scr_y < 0) continue;
//				if (scr_x > 640-34 || scr_y > 480-38) continue;

				if (map[x][y] == 2)
					sur_maptiles->put_screen(scr_x, scr_y, powerup_spr+1);
				else
					sur_maptiles->put_screen(scr_x, scr_y, map[x][y]);
			}
		}

		CL_Iterator<GameObject> object_counter(objects);
		while (object_counter.next() != NULL)
		{
			object_counter()->show(center_x, center_y);
		}

		float time_elapsed2 = time_elapsed;
		while (time_elapsed2 > 0)
		{
			float turn_time = (time_elapsed2>0.05) ? 0.05 : time_elapsed2;
			while (object_counter.next() != NULL)
			{
				if (object_counter()->turn(turn_time) == false)
				{
					if (object_counter() == player) player = NULL;
					delete object_counter();
					object_counter.remove();
				}
			}
			
			time_elapsed2 -= 0.05;
		}

		if (player == NULL)
		{
			fnt_clansoft->print_center(320, 320, "YOU ARE DEAD");
			fnt_clansoft->print_center(320, 360, "Press space to restart");

			if (welcome_shown == true && blowups.get_num_items() < 10)
			{
				blowups.add(new FontBlowUp("Welcome to the ClanSoft PacMan Game", 320, 20, fnt_clansoft));
				welcome_shown = false;
			}
		}
		else
		{
			welcome_shown = true;
			fnt_clansoft->print_center(320, 20, "Welcome to the ClanSoft PacMan Game");
		}

		CL_String score_string("");
		score_string += score;
		score_string += " bonus bananas";
		fnt_clansoft->print_right(620, 420, score_string.get_string());

		CL_Iterator<FontBlowUp> blowup_counter(blowups);
		while (blowup_counter.next() != NULL)
		{
			if (blowup_counter()->show(time_elapsed) == false)
			{
				delete blowup_counter();
				blowup_counter.remove();
			}
		}

		CL_Display::flip_display();	

		if (nof_eggs == 0) 		// level completed
		{
			init_map(0);
       			
			CL_Iterator<GameObject> object_counter(objects);
                	while (object_counter.next() != NULL)
                	{
    				delete object_counter();
                                object_counter.remove();        
                	}
	             
			init_objects();
		}

		if (CL_System::keep_alive()) break;
	}

	float delta_time = (CL_System::get_time()-start_time)/(float) 1000;
	cout << "tid nu: " << CL_System::get_time() << endl;
	cout << "start tid: " << start_time << endl;
	cout << "Frames: " << cur_frame << endl;
	cout << "Time elapsed: " << delta_time << endl;
	cout << "Framerate: " << cur_frame / delta_time << endl;
}

void PacManApplication::deinit()
{
	delete sur_maptiles;
	delete sur_pacman;
	delete sur_ghost;
	delete sur_bar;
	delete fnt_clansoft;
	delete sfx_namnam;
}
